import logging
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional, Tuple
from uuid import UUID

from kestrel.display import GraphletExplanation
from kestrel.exceptions import DataSourceError, InvalidDataSource
from kestrel.interface import DatasourceInterface
from kestrel.ir.graph import IRGraphEvaluable
from kestrel.ir.instructions import (
    DataSource,
    Filter,
    Instruction,
    Return,
    SolePredecessorTransformingInstruction,
    SourceInstruction,
    TransformingInstruction,
    Variable,
)
from kestrel.mapping.data_model import translate_dataframe
from kestrel_interface_opensearch.config import load_config
from kestrel_interface_opensearch.ossql import OpenSearchTranslator
from opensearchpy import OpenSearch
from pandas import DataFrame, concat

_logger = logging.getLogger(__name__)


def _jdbc2df(schema: dict, datarows: dict) -> DataFrame:
    """Convert a JDBC query result response to a DataFrame"""
    columns = [c.get("alias", c["name"]) for c in schema]
    return DataFrame(datarows, columns=columns)


def read_sql(sql: str, conn: OpenSearch, dmm: Optional[dict] = None) -> DataFrame:
    """Execute `sql` and return the results as a DataFrame, a la pandas.read_sql"""
    # https://opensearch.org/docs/latest/search-plugins/sql/sql-ppl-api/#query-api
    body = {
        # Temporarily comment out fetch_size due to https://github.com/opensearch-project/sql/issues/2579
        # FIXME: "fetch_size": 10000,  # Should we make this configurable?
        "query": sql,
    }
    query_resp = conn.http.post("/_plugins/_sql?format=jdbc", body=body)
    status = query_resp.get("status", 500)
    if status != 200:
        raise DataSourceError(f"OpenSearch query returned {status}")
    _logger.debug(
        "total=%d size=%d rows=%d",
        query_resp["total"],
        query_resp["size"],
        len(query_resp["datarows"]),
    )

    # Only the first page contains the schema
    # https://opensearch.org/docs/latest/search-plugins/sql/sql-ppl-api/#paginating-results
    schema = query_resp["schema"]
    dfs = []
    done = False
    while not done:
        df = _jdbc2df(schema, query_resp["datarows"])
        if dmm is not None:
            # Need to use Data Model Map to do results translation
            dfs.append(translate_dataframe(df, dmm))
        else:
            dfs.append(df)
        cursor = query_resp.get("cursor")
        if not cursor:
            break
        query_resp = conn.http.post(
            "/_plugins/_sql?format=jdbc", body={"cursor": cursor}
        )

    # Merge all pages together
    return concat(dfs)


class OpenSearchInterface(DatasourceInterface):
    def __init__(
        self,
        serialized_cache_catalog: Optional[str] = None,
        session_id: Optional[UUID] = None,
    ):
        super().__init__(serialized_cache_catalog, session_id)
        self.config = load_config()
        self.schemas: dict = {}  # Schema per table (index)
        self.conns: dict = {}  # Map of conn name -> connection
        for info in self.config.datasources.values():
            name = info.connection
            if name not in self.conns:
                conn = self.config.connections[name]
                client = OpenSearch(
                    [conn.url],
                    http_auth=(conn.auth.username, conn.auth.password),
                    verify_certs=conn.verify_certs,
                )
                self.conns[name] = client

    @staticmethod
    def schemes() -> Iterable[str]:
        return ["opensearch"]

    def get_datasources(self) -> List[str]:
        return list(self.config.datasources)

    def get_storage_of_datasource(datasource: str) -> str:
        """Get the storage name of a given datasource"""
        if datasource not in self.config.datasources:
            raise InvalidDataSource(datasource)
        return self.config.datasources[datasource].connection

    def store(
        self,
        instruction_id: UUID,
        data: DataFrame,
    ):
        raise NotImplementedError("OpenSearchInterface.store")  # TEMP

    def evaluate_graph(
        self,
        graph: IRGraphEvaluable,
        cache: MutableMapping[UUID, Any],
        instructions_to_evaluate: Optional[Iterable[Instruction]] = None,
    ) -> Mapping[UUID, DataFrame]:
        mapping = {}
        if not instructions_to_evaluate:
            instructions_to_evaluate = graph.get_sink_nodes()
        for instruction in instructions_to_evaluate:
            translator, datasource = self._evaluate_instruction_in_graph(
                graph, instruction
            )
            # TODO: may catch error in case evaluation starts from incomplete SQL
            sql = translator.result()
            _logger.debug("SQL query generated: %s", sql)
            ds = self.config.datasources[datasource]
            conn = self.config.connections[ds.connection]
            client = OpenSearch(
                [conn.url],
                http_auth=(conn.auth.username, conn.auth.password),
                verify_certs=conn.verify_certs,
            )
            mapping[instruction.id] = read_sql(
                sql, client, translator.from_ocsf_map[translator.entity]
            )
            client.close()
        return mapping

    def explain_graph(
        self,
        graph: IRGraphEvaluable,
        instructions_to_explain: Optional[Iterable[Instruction]] = None,
    ) -> Mapping[UUID, GraphletExplanation]:
        mapping = {}
        if not instructions_to_explain:
            instructions_to_explain = graph.get_sink_nodes()
        for instruction in instructions_to_explain:
            translator = self._evaluate_instruction_in_graph(graph, instruction)
            dep_graph = graph.duplicate_dependent_subgraph_of_node(instruction)
            graph_dict = dep_graph.to_dict()
            query_stmt = translator.result()
            mapping[instruction.id] = GraphletExplanation(graph_dict, query_stmt)
        return mapping

    def _evaluate_instruction_in_graph(
        self,
        graph: IRGraphEvaluable,
        instruction: Instruction,
    ) -> Tuple[OpenSearchTranslator, str]:
        _logger.debug("instruction: %s", str(instruction))
        translator = None
        datasource = None
        if isinstance(instruction, TransformingInstruction):
            trunk, _r2n = graph.get_trunk_n_branches(instruction)
            translator, datasource = self._evaluate_instruction_in_graph(graph, trunk)

            if isinstance(instruction, SolePredecessorTransformingInstruction):
                if isinstance(instruction, Return):
                    pass
                elif isinstance(instruction, Variable):
                    pass
                else:
                    translator.add_instruction(instruction)

            elif isinstance(instruction, Filter):
                translator.add_instruction(instruction)

            else:
                raise NotImplementedError(f"Unknown instruction type: {instruction}")

        elif isinstance(instruction, SourceInstruction):
            if isinstance(instruction, DataSource):
                datasource = instruction.datasource
                ds = self.config.datasources[instruction.datasource]
                schema = self.get_schema(instruction.datasource)
                translator = OpenSearchTranslator(
                    ds.timestamp_format,
                    ds.timestamp,
                    ds.index_pattern,
                    ds.data_model_map,
                    schema,
                )
            else:
                raise NotImplementedError(f"Unhandled instruction type: {instruction}")

        return translator, datasource

    def _get_client_for_datasource(self, datasource: str) -> OpenSearch:
        ds = self.config.datasources[datasource]
        conn = ds.connection
        _logger.debug(
            "Fetching schema for %s from %s",
            datasource,
            self.config.connections[conn].url,
        )
        return self.conns[conn], ds.index_pattern

    def get_schema(self, datasource: str) -> dict:
        client, index = self._get_client_for_datasource(datasource)
        if index not in self.schemas:
            df = read_sql(f"DESCRIBE TABLES LIKE {index}", client)
            self.schemas[datasource] = (
                df[["TYPE_NAME", "COLUMN_NAME"]]
                .set_index("COLUMN_NAME")
                .T.to_dict("records")[0]
            )
            _logger.debug("%s schema:\n%s", datasource, self.schemas[datasource])
        return self.schemas[datasource]



import os

import yaml
from kestrel_interface_opensearch.config import (PROFILE_PATH_ENV_VAR,
                                                 Connection, load_config)


def test_load_config(tmp_path):
    config = {
        "connections": {
            "localhost": {
                "url": "https://localhost:9200",
                "verify_certs": False,
                "auth": {
                    "username": "admin",
                    "password": "admin",
                }
            },
            "some-cloud-thing": {
                "url": "https://www.example.com:9200",
                "verify_certs": True,
                "auth": {
                    "username": "hunter",
                    "password": "super_secret",
                }
            }
        },
        "datasources": {
            "some_ds": {
                "connection": "some-cloud-thing",
                "index_pattern": "logs-*",
                "timestamp": "@timestamp",
                "timestamp_format": "%Y-%m-%d %H:%M:%S.%f",
                "data_model_map": "mapping.yaml"
            }
        }
    }
    map_file = tmp_path / "mapping.yaml"
    with open(map_file, 'w') as fp:
        fp.write("some.field: other.field\n")
    config_file = tmp_path / "opensearch.yaml"
    with open(config_file, 'w') as fp:
        yaml.dump(config, fp)
    os.environ[PROFILE_PATH_ENV_VAR] = str(config_file)
    read_config = load_config()
    conn: Connection = read_config.connections["localhost"]
    assert conn.url == config["connections"]["localhost"]["url"]
    assert read_config.connections["localhost"].url == config["connections"]["localhost"]["url"]
    assert read_config.datasources["some_ds"].index_pattern == config["datasources"]["some_ds"]["index_pattern"]
    assert read_config.datasources["some_ds"].data_model_map["some.field"] == "other.field"



from datetime import datetime

import pytest
from dateutil import parser
from kestrel_interface_opensearch.ossql import OpenSearchTranslator

from kestrel.exceptions import UnsupportedOperatorError
from kestrel.ir.filter import (ExpOp, IntComparison, ListComparison, ListOp,
                               MultiComp, NumCompOp, StrComparison, StrCompOp,
                               TimeRange)
from kestrel.ir.instructions import (Filter, Limit, Offset, ProjectAttrs,
                                     ProjectEntity, Sort)

TIMEFMT = '%Y-%m-%dT%H:%M:%S.%fZ'


# A much-simplified test mapping
data_model_map = {
    "process": {
        "cmd_line": "CommandLine",
        "file": {
            "path": "Image",
            # "name": [
            #     {
            #         "native_field": "Image",
            #         "native_value": "basename",
            #         "ocsf_op": "LIKE",
            #         "ocsf_value": "endswith"
            #     }
            # ]
        },
        "pid": "ProcessId",
        "parent_process": {
            "pid": "ParentProcessId",
        },
    },
}

schema = {
    "CommandLine": "text",
    "Image": "text",
    "ProcessId": "text",
    "ParentProcessId": "text",
}


def _dt(timestr: str) -> datetime:
    return parser.parse(timestr)


def _remove_nl(s):
    return s.replace('\n', '')


@pytest.mark.parametrize(
    "iseq, sql", [
        # Try a simple filter
        ([Filter(IntComparison('foo', NumCompOp.GE, 0))],
         "SELECT {} FROM my_table WHERE `foo` >= 0"),
        # Try a simple filter with sorting
        ([Filter(IntComparison('foo', NumCompOp.GE, 0)), Sort('bar')],
         "SELECT {} FROM my_table WHERE `foo` >= 0 ORDER BY `bar` DESC"),
        # Simple filter plus time range
        ([Filter(IntComparison('foo', NumCompOp.GE, 0), timerange=TimeRange(_dt('2023-12-06T08:17:00Z'), _dt('2023-12-07T08:17:00Z')))],
         "SELECT {} FROM my_table WHERE `foo` >= 0 AND `timestamp` >= '2023-12-06T08:17:00.000000Z' AND `timestamp` < '2023-12-07T08:17:00.000000Z'"),
        # Add a limit and projection
        ([Limit(3), ProjectAttrs(['foo', 'bar', 'baz']), Filter(StrComparison('foo', StrCompOp.EQ, 'abc'))],
         "SELECT `foo`, `bar`, `baz` FROM my_table WHERE `foo` = 'abc' LIMIT 3"),
        # Same as above but reverse order
        ([Filter(StrComparison('foo', StrCompOp.EQ, 'abc')), ProjectAttrs(['foo', 'bar', 'baz']), Limit(3)],
         "SELECT `foo`, `bar`, `baz` FROM my_table WHERE `foo` = 'abc' LIMIT 3"),
        ([Filter(ListComparison('foo', ListOp.NIN, ['abc', 'def']))],
         "SELECT {} FROM my_table WHERE `foo` NOT IN ('abc', 'def')"),
        ([Filter(MultiComp(ExpOp.OR, [IntComparison('foo', NumCompOp.EQ, 1), IntComparison('bar', NumCompOp.EQ, 1)]))],
         "SELECT {} FROM my_table WHERE `foo` = 1 OR `bar` = 1"),
        ([Filter(MultiComp(ExpOp.AND, [IntComparison('foo', NumCompOp.EQ, 1), IntComparison('bar', NumCompOp.EQ, 1)]))],
         "SELECT {} FROM my_table WHERE `foo` = 1 AND `bar` = 1"),
        ([Limit(1000), Offset(2000)],
         "SELECT {} FROM my_table LIMIT 2000, 1000"),
        # Test entity projection
        ([Limit(3), Filter(StrComparison('cmd_line', StrCompOp.EQ, 'foo bar')), ProjectEntity('process', 'process')],
         "SELECT {} FROM my_table WHERE `CommandLine` = 'foo bar' LIMIT 3"),
    ]
)
def test_opensearch_translator(iseq, sql):
    if ProjectEntity in {type(i) for i in iseq}:
        cols = '`CommandLine` AS `cmd_line`, `Image` AS `file.path`, `ProcessId` AS `pid`, `ParentProcessId` AS `parent_process.pid`'
    else:
        cols = '`CommandLine` AS `process.cmd_line`, `Image` AS `process.file.path`, `ProcessId` AS `process.pid`, `ParentProcessId` AS `process.parent_process.pid`'
    trans = OpenSearchTranslator(TIMEFMT, "timestamp", "my_table", data_model_map, schema)
    for i in iseq:
        trans.add_instruction(i)
    result = trans.result()
    assert _remove_nl(str(result)) == sql.format(cols)


@pytest.mark.parametrize(
    "instruction", [
        Filter(StrComparison('foo', StrCompOp.MATCHES, '.*abc.*')),
        Filter(StrComparison('foo', StrCompOp.NMATCHES, '.*abc.*')),
    ]
)
def test_opensearch_translator_unsupported(instruction):
    trans = OpenSearchTranslator(TIMEFMT, "timestamp", "my_table", data_model_map, schema)
    with pytest.raises(UnsupportedOperatorError):
        trans.add_instruction(instruction)
        _ = trans.result()



import logging
from dataclasses import dataclass
from typing import Dict, Optional

from kestrel.config.utils import (
    CONFIG_DIR_DEFAULT,
    load_kestrel_config,
    load_user_config,
)
from kestrel.exceptions import InterfaceNotConfigured
from kestrel.mapping.data_model import (
    check_entity_identifier_existence_in_mapping,
    load_default_mapping,
)
from mashumaro.mixins.json import DataClassJSONMixin

PROFILE_PATH_DEFAULT = CONFIG_DIR_DEFAULT / "opensearch.yaml"
PROFILE_PATH_ENV_VAR = "KESTREL_OPENSEARCH_CONFIG"

_logger = logging.getLogger(__name__)


@dataclass
class Auth:
    username: str
    password: str


@dataclass
class Connection(DataClassJSONMixin):
    url: str
    auth: Auth
    verify_certs: bool = True

    def __post_init__(self):
        self.auth = Auth(**self.auth)


@dataclass
class DataSource(DataClassJSONMixin):
    connection: str
    index_pattern: str
    timestamp: str
    timestamp_format: str
    data_model_map: Optional[Dict] = None
    entity_identifier: Optional[Dict] = None

    def __post_init__(self):
        if not self.data_model_map:
            # Default to the built-in ECS mapping
            self.data_model_map = load_default_mapping("ecs")

        kestrel_config = load_kestrel_config()
        check_entity_identifier_existence_in_mapping(
            self.data_model_map,
            kestrel_config["entity_identifier"],
            "opensearch interface",
        )


@dataclass
class Config(DataClassJSONMixin):
    connections: Dict[str, Connection]
    datasources: Dict[str, DataSource]

    def __post_init__(self):
        self.connections = {k: Connection(**v) for k, v in self.connections.items()}
        self.datasources = {k: DataSource(**v) for k, v in self.datasources.items()}


def load_config():
    try:
        interface_config = Config(
            **load_user_config(PROFILE_PATH_ENV_VAR, PROFILE_PATH_DEFAULT)
        )
        return interface_config
    except TypeError:
        raise InterfaceNotConfigured()



import logging
from functools import reduce
from typing import Optional, Union

from kestrel.exceptions import UnsupportedOperatorError
from kestrel.ir.filter import (
    BoolExp,
    ExpOp,
    FComparison,
    ListOp,
    MultiComp,
    NumCompOp,
    StrComparison,
    StrCompOp,
)
from kestrel.ir.instructions import (
    Filter,
    Instruction,
    Limit,
    Offset,
    ProjectAttrs,
    ProjectEntity,
    Sort,
    SortDirection,
)
from kestrel.mapping.data_model import (
    translate_comparison_to_native,
    translate_projection_to_native,
)
from typeguard import typechecked

_logger = logging.getLogger(__name__)


Value = Union[
    int,
    float,
    str,
    list,
]


@typechecked
def _and(lhs: str, rhs: Value) -> str:
    return " AND ".join((lhs, rhs))


@typechecked
def _or(lhs: str, rhs: Value) -> str:
    return " OR ".join((lhs, rhs))


# SQL comparison operator functions
comp2func = {
    NumCompOp.EQ: "=",
    NumCompOp.NEQ: "<>",
    NumCompOp.LT: "<",
    NumCompOp.LE: "<=",
    NumCompOp.GT: ">",
    NumCompOp.GE: ">=",
    StrCompOp.EQ: "=",
    StrCompOp.NEQ: "<>",
    StrCompOp.LIKE: "LIKE",
    StrCompOp.NLIKE: "NOT LIKE",
    # UNSUPPORTED BY OpenSearch SQL: StrCompOp.MATCHES: "REGEXP",
    # UNSUPPORTED BY OpenSearch SQL: StrCompOp.NMATCHES: "NOT REGEXP",
    ListOp.IN: "IN",
    ListOp.NIN: "NOT IN",
}


def _format_value(value):
    if isinstance(value, str):
        # Need to quote string values
        value = f"'{value}'"
    elif isinstance(value, list):
        # SQL uses parens for lists
        value = tuple(value)
    return value


@typechecked
class OpenSearchTranslator:
    def __init__(
        self,
        timefmt: str,
        timestamp: str,
        select_from: str,
        data_model_map: dict,
        schema: dict,
    ):
        # Time format string for datasource
        self.timefmt = timefmt

        # Primary timestamp field in target table
        self.timestamp = timestamp

        # Query clauses
        self.table: str = select_from
        self.filt: Optional[Filter] = None
        self.projection_base_field: Optional[str] = None
        self.project: Optional[ProjectAttrs] = None
        self.limit: int = 0
        self.offset: int = 0
        self.order_by: str = ""
        self.sort_dir = SortDirection.DESC

        # Data model mapping: should be ocsf -> native
        self.from_ocsf_map = data_model_map

        # Index "schema" (field name -> type)
        self.schema = schema

    @typechecked
    def _render_comp(self, comp: FComparison) -> str:
        prefix = (
            f"{self.projection_base_field}."
            if (self.projection_base_field and comp.field != self.timestamp)
            else ""
        )
        ocsf_field = f"{prefix}{comp.field}"
        comps = translate_comparison_to_native(
            self.from_ocsf_map, ocsf_field, comp.op, comp.value
        )
        try:
            comps = [f"`{f}` {comp2func[o]} {_format_value(v)}" for f, o, v in comps]
            conj = " OR ".join(comps)
            result = conj if len(comps) == 1 else f"({conj})"
        except KeyError:
            raise UnsupportedOperatorError(
                comp.op.value
            )  # FIXME: need to report the mapped op, not the original
        return result

    @typechecked
    def _render_multi_comp(self, comps: MultiComp) -> str:
        op = _and if comps.op == ExpOp.AND else _or
        return reduce(op, map(self._render_comp, comps.comps))

    @typechecked
    def _render_exp(self, exp: BoolExp) -> str:
        if isinstance(exp.lhs, BoolExp):
            lhs = self._render_exp(exp.lhs)
        elif isinstance(exp.lhs, MultiComp):
            lhs = self._render_multi_comp(exp.lhs)
        else:
            lhs = self._render_comp(exp.lhs)
        if isinstance(exp.rhs, BoolExp):
            rhs = self._render_exp(exp.rhs)
        elif isinstance(exp.rhs, MultiComp):
            rhs = self._render_multi_comp(exp.rhs)
        else:
            rhs = self._render_comp(exp.rhs)
        return _and(lhs, rhs) if exp.op == ExpOp.AND else _or(lhs, rhs)

    @typechecked
    def _render_filter(self) -> Optional[str]:
        if not self.filt:
            return None
        if self.filt.timerange.start:
            # Convert the timerange to the appropriate pair of comparisons
            start_comp = StrComparison(
                self.timestamp, ">=", self.filt.timerange.start.strftime(self.timefmt)
            )
            stop_comp = StrComparison(
                self.timestamp, "<", self.filt.timerange.stop.strftime(self.timefmt)
            )
            # AND them together
            time_exp = BoolExp(start_comp, ExpOp.AND, stop_comp)
            # AND that with any existing filter expression
            exp = BoolExp(self.filt.exp, ExpOp.AND, time_exp)
        else:
            exp = self.filt.exp
        if isinstance(exp, BoolExp):
            comp = self._render_exp(exp)
        elif isinstance(exp, MultiComp):
            comp = self._render_multi_comp(exp)
        else:
            comp = self._render_comp(exp)
        return comp

    def add_Filter(self, filt: Filter) -> None:
        # Just save filter and compile it later
        # Probably need the entity projection set first
        self.filt = filt

    def add_ProjectAttrs(self, proj: ProjectAttrs) -> None:
        # Just save projection and compile it later
        self.project = proj

    def _render_proj(self):
        """Get a list of native cols to project with their OCSF equivalents as SQL aliases"""
        projection = self.project.attrs if self.project else None
        name_pairs = translate_projection_to_native(
            self.from_ocsf_map, self.projection_base_field, projection
        )
        proj = [
            f"`{k}` AS `{v}`" if k != v else f"`{k}`"
            for k, v in name_pairs
            if k in self.schema  # Ignore mapped attrs the index doesn't have
        ]
        if not proj:
            # If this is still empty, then the attr projection must be for attrs "outside" to entity projection?
            proj = [f"`{attr}`" for attr in self.project.attrs]
        _logger.debug("Set projection to %s", proj)
        return proj

    def add_ProjectEntity(self, proj: ProjectEntity) -> None:
        self.projection_base_field = proj.ocsf_field
        _logger.debug("Set base entity to '%s'", self.projection_base_field)

    def add_Limit(self, lim: Limit) -> None:
        self.limit = lim.num

    def add_Offset(self, offset: Offset) -> None:
        self.offset = offset.num

    def add_Sort(self, sort: Sort) -> None:
        self.order_by = sort.attribute
        self.sort_dir = sort.direction

    def add_instruction(self, i: Instruction) -> None:
        inst_name = i.instruction
        method_name = f"add_{inst_name}"
        try:
            method = getattr(self, method_name)
        except AttributeError as e:
            raise NotImplementedError(f"OpenSearchTranslator.{method_name}")
        method(i)

    def result(self) -> str:
        stages = ["SELECT"]
        cols = ", ".join(self._render_proj())
        stages.append(f"{cols}")
        stages.append(f"FROM {self.table}")
        where = self._render_filter()
        if where:
            stages.append(f"WHERE {where}")
        if self.order_by:
            stages.append(f"ORDER BY `{self.order_by}` {self.sort_dir.value}")
        if self.limit:
            # https://opensearch.org/docs/latest/search-plugins/sql/sql/basic/#limit
            if self.offset:
                stages.append(f"LIMIT {self.offset}, {self.limit}")
            else:
                stages.append(f"LIMIT {self.limit}")
        sql = " ".join(stages)
        _logger.debug("SQL: %s", sql)
        return sql
